/*
 Do not modify, auto-generated by script

 Copyright 2019 Alain Dargelas

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

/*
 * File:   Serializer_save.cpp
 * Author:
 *
 * Created on December 14, 2019, 10:03 PM
 */
#include <uhdm/Serializer.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#if defined(_MSC_VER)
  #include <io.h>
#else
  #include <unistd.h>
#endif

#include <iostream>
#include <string>
#include <vector>

#include <capnp/message.h>
#include <capnp/serialize-packed.h>

#include "UHDM.capnp.h"
#include <uhdm/containers.h>
#include <uhdm/uhdm.h>
#include <uhdm/uhdm_types.h>

#include "uhdm/config.h"

namespace UHDM {
inline static uint32_t GetId(const BaseClass* p, const Serializer::IdMap &idMap) {
  auto it = idMap.find(p);
  return (it == idMap.end()) ? Serializer::kBadIndex : it->second;
}

struct Serializer::SaveAdapter {
  void operator()(const BaseClass *const obj, Serializer *const serializer, const IdMap& idMap, Any::Builder builder) const {
    if (obj->VpiParent() != nullptr) {
      ::ObjIndexType::Builder vpiParentBuilder = builder.getVpiParent();
      vpiParentBuilder.setIndex(GetId(obj->VpiParent(), idMap));
      vpiParentBuilder.setType(static_cast<uint32_t>(obj->VpiParent()->UhdmType()));
    }
    builder.setVpiFile((RawSymbolId)obj->GetSerializer()->symbolMaker.Make(obj->VpiFile()));
    builder.setVpiLineNo(obj->VpiLineNo());
    builder.setVpiColumnNo(obj->VpiColumnNo());
    builder.setVpiEndLineNo(obj->VpiEndLineNo());
    builder.setVpiEndColumnNo(obj->VpiEndColumnNo());
    builder.setUhdmId(obj->UhdmId());
  }

<CAPNP_SAVE_ADAPTERS>

  template<typename T, typename U, typename = typename std::enable_if<std::is_base_of<BaseClass, T>::value>::type>
  void operator()(const FactoryT<T>& factory, Serializer* serializer, const Serializer::IdMap& idMap,
                  typename ::capnp::List<U>::Builder builder) const {
    uint32_t index = 0;
    for (const T* obj : factory.objects_)
      operator()(obj, serializer, idMap, builder[index++]);
  }
};

void Serializer::Save(const std::filesystem::path& filepath) {
    Save(filepath.string());
}

void Serializer::Save(const std::string& filepath) {
  if (m_enableGC) GarbageCollect();

  uint32_t index = 0;
  const std::string file = filepath;
  const int32_t fileid = open(file.c_str(), O_CREAT | O_WRONLY | O_BINARY, S_IRWXU);

  ::capnp::MallocMessageBuilder message;
  UhdmRoot::Builder cap_root = message.initRoot<UhdmRoot>();
  cap_root.setVersion(kVersion);
  cap_root.setObjectId(m_objId);

  ::capnp::List<Design>::Builder designs = cap_root.initDesigns(designMaker.objects_.size());
  index = 0;
  for (auto design : designMaker.objects_) {
    designs[index].setVpiName((RawSymbolId)design->GetSerializer()->symbolMaker.Make(design->VpiName()));
    index++;
  }

  const IdMap idMap = AllObjects();
  SaveAdapter adapter;
  <CAPNP_SAVE>

  // Save the symbols after all save function have been invoked, some symbols are made doing so (VpiFullName)
  // This is not ideal.
  // Ideally, the save should not include the hierarchical nets that can be recreated on the fly.
  // Something broke this mechanism that saved a lot of memory/disk space.
  // Until that is repaired we go for the more disk-hungry and memory hungry method which gives correct results.
  ::capnp::List<::capnp::Text>::Builder symbols = cap_root.initSymbols(symbolMaker.m_id2SymbolMap.size());
  index = 0;
  for (const auto& symbol : symbolMaker.m_id2SymbolMap) {
    symbols.set(index, symbol.c_str());
    index++;
  }

  writePackedMessageToFd(fileid, message);
  close(fileid);
}
}  // namespace UHDM
